__declspec(dllimport) void __cdecl start(void(__cdecl*)());
__declspec(dllimport) void __stdcall start(void(__stdcall*)());
__declspec(dllimport) void __fastcall start(void(__fastcall*)());
__declspec(dllimport) void __thiscall start(void(__thiscall*)());
