__attribute__((annotate("an annotation"))) void __cdecl start(
  void(__cdecl*)());
__attribute__((annotate("an annotation"))) void __stdcall start(
  void(__stdcall*)());
__attribute__((annotate("an annotation"))) void __fastcall start(
  void(__fastcall*)());
__attribute__((annotate("an annotation"))) void __thiscall start(
  void(__thiscall*)());
