class __attribute__((deprecated(
  "Deprecated: Additional information about the deprecation of 'start'\
which spreads over multiple lines <this is more infomation> can be found \"online\" ")))
start
{
};
