__attribute__((deprecated("Function Deprecated"))) void __cdecl start(
  void(__cdecl*)());
__attribute__((deprecated(
  "stdcall Function Deprecated"))) void __stdcall start(void(__stdcall*)());
__attribute__((deprecated(
  "fastcall Function Deprecated"))) void __fastcall start(void(__fastcall*)());
__attribute__((deprecated(
  "thiscall Function Deprecated"))) void __thiscall start(void(__thiscall*)());
