namespace start {
typedef char t_Char;
typedef signed char t_SChar;
typedef unsigned char t_UChar;
typedef short t_Short;
typedef unsigned short t_UShort;
typedef int t_Int;
typedef unsigned int t_UInt;
typedef long t_Long;
typedef unsigned long t_ULong;
typedef long long t_LongLong;
typedef unsigned long long t_ULongLong;
typedef float t_Float;
typedef double t_Double;
}
